#include <iostream>
#include "hyperEA3_parse.h"

namespace hea {

bool parse_set_content(const char **str, const int n, int &set_number) {
	const char *pom = *str;

	int tmp_val = 0;
	while (*pom && *pom != '}') {
		// parse one elemet
		// skip ws
		while (*pom && (*pom == ' ' || *pom == '\t') )
			pom++;

		if (*pom=='}') {
			continue;
		}
		else {
			if (*pom == '0' || *pom =='1' || (*pom >='a' && *pom <='z')) {
				switch (*pom) {
					case '0':
						tmp_val |= 1;
						break;
					case '1':
						tmp_val |= (1<< (n-1));
						break;

					default:
						tmp_val |= (1<< (int)(*pom-'a' + 1));
						break;
				}
				pom++;
			}
			else {
				// unexpected character
				std::cout << "Unexpected character '" << *pom << "'" << std::endl;
				return false;
			}

			// skip ws
			while (*pom && (*pom == ' ' || *pom == '\t') )
				pom++;
			if (*pom == ',')
				pom++;
		}
	}

	if (*pom == '}') {
		set_number = tmp_val;
		//pom++;
		*str = pom;
		return true;
	}
	
	std::cout << "HERE *pom=='" << *pom << "'" << std::endl;
	return false;
}

bool parse_set(const char **str, const int n, int &set_number) {
	const char *pom = *str;
	// skip ws
	while (*pom && (*pom == ' ' || *pom == '\t') )
		pom++;

	if (*pom != '{') {
		// missing openning brace {
		std::cout << "Missing {" << std::endl;
		return false;
	}
	else
		pom++;

	if (!parse_set_content(&pom, n, set_number)) {
		std::cout << "Parse set content error" << std::endl;
		return false;
	}

	// skip ws
	while (*pom && (*pom == ' ' || *pom == '\t') )
		pom++;
	
	if (*pom != '}') {
		// missing closing brace }
		std::cout << "Missing }" << std::endl;
		return false;
	}
	else
		pom++;

	*str = pom;
	return true;
} 

bool parse_line(const char **str, const int n, int *line_nums) {
	const char *pom = *str;
	int idx = 0;

	while (*pom && *pom != '\n' && *pom !='\r') {
		

		int set_number;
		if (parse_set(&pom, n, set_number)) {
			//cout << set_number << " ";
			line_nums[idx] = set_number;
			idx++;
			
		}
		else {
			std::cout << "Error parsing set" << std::endl;
			return false;
		}

		// skip ws
		while (*pom && (*pom == ' ' || *pom == '\t') )
			pom++;

		if (*pom==',')
			pom++;

		// skip ws
		while (*pom && (*pom == ' ' || *pom == '\t') )
			pom++;
	}

	//cout <<endl;
	return true;
}

bool parse_op(std::ifstream &ifs, const int n, const std::set<int>* (&op)[MAX][MAX], const std::set<int>* setCandidates) {
	char buffer[1024];
	const char *pom = buffer;
	int nthline = 0;
	int line_nums[MAX*2];
	
	while (ifs.good() && nthline < n) {
		ifs.getline(buffer, 1024);

		if (ifs.good()) {
			if (parse_line(&pom, n, line_nums)) {
				for (int i = 0; i < n; i++)
					op[nthline][i] = &setCandidates[ line_nums[i] ];
			}
			else
				// error parsing line
				return false;
		
			nthline++;
		}
	}
	return true;

	
}

bool parse_complement(std::ifstream &ifs, const int n, int (&complement)[MAX]) {
	char buffer[1024];

	// skip first line
	if (ifs.good())
		ifs.getline(buffer, 1024);
	else
		// cannot read stream
		return false;

	if (ifs.good())
		ifs.getline(buffer, 1024);
	else
		// cannot read stream
		return false;
	
	//cout << "Buffer: \"" << buffer << "\"" << endl;


	char *pom = &buffer[0];
	int idx = 0;
	int tmp_val = 0;

	while (*pom && idx < n) {
		//cout << "'" << *pom << "'" ;
		// skip ws
		while (*pom && (*pom == ' ' || *pom == '\t') )
			pom++;
		//cout << "'" << *pom << "'" << endl;

		if (*pom == '0' || *pom =='1' || (*pom >='a' && *pom <='z')) {
			switch (*pom) {
				case '0':
					tmp_val = 0;
					break;
				case '1':
					tmp_val = n - 1;
					break;

				default:
					tmp_val = (int)(*pom-'a' + 1);
					break;
			}
			//cout << tmp_val << endl;
			complement[idx] = tmp_val;
			idx++;
			pom++;
		}
		else
			// Unknown character
			return false;
		
		
	}

	if (idx != n)
		// insufficient number of elements read
		return false;

	return true;
}

bool parse_model(std::ifstream &ifs, const int n, const std::set<int>* (&op)[MAX][MAX], int (&complement)[MAX], const std::set<int> *setCandidates) {
	if (parse_complement(ifs, n, complement)) {
		if (parse_op(ifs, n, op, setCandidates))
			return true;
		else
			// error parsing op
			return false;

	}
	// error parsing complement
	return false;
}

}
