#include <iostream>
#include <set>
#include <vector>
#include <map>
#include <stdlib.h>

#include "consts.h"
#include "permutation.h"
#include "hyperEA3_common.h"
#include "candidate.h"

using namespace std;


struct Relation {
	int n;
	int relation[MAX][MAX];

	Relation(const int _n, const int (&_relation)[MAX][MAX]) {
		n = _n;

		for (int i = 0; i < _n; i++)	
		for (int j = 0; j < _n; j++)	

			relation[i][j] = _relation[i][j];
	}
};


struct Zoznamy {
	set<pair<int, int> > one_pairs;
	set<pair<int, int> > other_pairs;

	Zoznamy() {
	}

	Zoznamy( const set<pair<int, int> > &_one_pairs, const set<pair<int, int> > &_other_pairs ) {
		for (set<pair<int, int> >::const_iterator it = _one_pairs.begin(); it != _one_pairs.end(); ++it) {
			one_pairs.insert(*it);
		}

		for (set<pair<int, int> >::const_iterator it = _other_pairs.begin(); it != _other_pairs.end(); ++it) {
			other_pairs.insert(*it);
		}
	}
};



bool are_rel_isomorphic_via_perm(const int n, const int (&permutation)[MAX], const int (&relation)[MAX][MAX], const Relation &relation2) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (relation[i][j] != relation2.relation[ permutation[i] ][ permutation[j] ])
			return false;

	return true;

}

void are_isomorphic_recursive(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], const Relation &relation2, bool &are_iso) {
	if (!are_iso) {
		if (level == n - 1) {

			if (are_rel_isomorphic_via_perm(n, permutation, relation, relation2))
				are_iso = true;

		}
		else {
			for (int i=level; i < n; i++) {
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
				
				are_isomorphic_recursive(level + 1, permutation, n, relation, relation2, are_iso);
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}

			}
		}
	}
}

bool are_isomorphic(const int n, const int (&relation)[MAX][MAX], const Relation &relation2) {
	bool are_iso = false;
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	are_isomorphic_recursive(0, permutation, n, relation, relation2, are_iso);

	return are_iso;
}

bool is_isomorphic(const int n, const int (&relation)[MAX][MAX], const vector<Relation> &relations) {
	/* first relation */
	if (relations.empty())
		return false;


	for (vector<Relation>::const_iterator it = relations.begin(); it != relations.end(); ++it) {
		if (are_isomorphic(n, relation, *it))
			return true;
	}

	return false;
}

bool is_double_negation_permutation(const int n, const int (&permutation)[MAX]) {
	for (int i = 0; i < n; i++)
		if (i != permutation [ permutation [ i ] ])
			return false;

	return true;
}



void admits_reverse_of_relation_recursive(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], bool &found) {
	if (!found) {
		if (level == n - 1) {
			bool is_valid  = true;
			for (int i = 0; i < n && is_valid; i++)
			for (int j = 0; j < n && is_valid; j++)
				if (i != j)
					if (relation[i][j] && !relation[permutation[j]][permutation[i]]) {
						/* wrong permutation */
						is_valid = false;
						break;
					}

			if (is_valid) {
				found = true;


#if 0
				/* test permutation */	
				for (int i = 0; i < n; i++)
					cout << permutation[i] << " ";
				cout << endl;
				cout << "-P-" << endl;
#endif
			}
		}
		else {
			for (int i=level; i < n; i++) {
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
				
				admits_reverse_of_relation_recursive(level + 1, permutation, n, relation, found);
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
					
			}
				
		}
	}
}


bool admits_reverse_of_relation(const int n, const int (&relation)[MAX][MAX]) {
	
	bool found = false;
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	admits_reverse_of_relation_recursive(0, permutation, n, relation, found); 

	return found;
	
}

void get_vector_of_orthosupplement_candidates(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], vector<hea::Permutation> &vector_ortho) {

	if (level == n - 1) {
		bool is_valid  = true;
		if (!is_double_negation_permutation(n, permutation))
			is_valid = false;

		for (int i = 0; i < n && is_valid; i++)
		for (int j = 0; j < n && is_valid; j++)
			if (i != j)
				if (relation[i][j] && !relation[permutation[j]][permutation[i]]) {
					/* wrong permutation */
					is_valid = false;
					break;
				}

		if (is_valid) {
			vector_ortho.push_back(hea::Permutation(n, permutation));


#if 0
			/* test permutation */	
			for (int i = 0; i < n; i++)
				cout << permutation[i] << " ";
			cout << endl;
			cout << "-P-" << endl;
#endif
		}
	}
	else {
		for (int i=level; i < n; i++) {
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
			
			get_vector_of_orthosupplement_candidates(level + 1, permutation, n, relation, vector_ortho);
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
				
		}
			
	}
}

/* mapping x-> x', such that x<=y, iff y'<=x' 
 *  returns true, if there is at least one such mapping
 */
bool get_vector_of_orthosupplement_candidates(const int n, const int (&relation)[MAX][MAX], vector<hea::Permutation> &vector_ortho) {
	
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	get_vector_of_orthosupplement_candidates(0, permutation, n, relation, vector_ortho); 

	return !vector_ortho.empty();
}


bool is_similar_pairs_via_perm(const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs, const int (&permutation)[MAX]) {
	//set<pair<int, int> > one_iso_pairs, other_iso_pairs;

	for (set<pair<int, int> >::const_iterator it_set = one_pairs.begin(); it_set != one_pairs.end(); ++it_set) {
		pair<int, int> iso_pair(permutation[it_set->first - 1] + 1, permutation[it_set->second - 1] + 1);
		if (zoznamy.one_pairs.find(iso_pair) == zoznamy.one_pairs.end())
			/* not found */
			return false;


	//	one_iso_pairs.insert(pair<int, int>(permutation[it_set->first], permutation[it_set->second]));	

	}

	for (set<pair<int, int> >::const_iterator it_set = other_pairs.begin(); it_set != other_pairs.end(); ++it_set) {
		pair<int, int> iso_pair(permutation[it_set->first - 1] + 1, permutation[it_set->second - 1] + 1);
		if (zoznamy.other_pairs.find(iso_pair) == zoznamy.other_pairs.end())
			/* not found */
			return false;


	//	one_iso_pairs.insert(pair<int, int>(permutation[it_set->first], permutation[it_set->second]));	

	}

	return true;
}

/* is  zoznamy isomorphic to one_pairs, other_pairs*/
void is_similar_pairs_recursive(int level, const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs, int (&permutation)[MAX], bool &is_similar) {
	if (!is_similar) {
		if (level == n - 1) {
/*
			for (int i = 0; i < n; i++)
				cout << permutation[i] << " ";
			cout << endl;
*/

			if (is_similar_pairs_via_perm(n, zoznamy, one_pairs, other_pairs, permutation)) {
/*
				cout << "Permutation: ";
				for (int i = 0; i < n; i++)
					cout << permutation[i] ;
				cout << endl;
*/
				is_similar = true;
			}

		}
		else {

			for (int i = level; i < n; i++) {
				if (i!=level) {
					int tmp = permutation[i];
					permutation[i] = permutation[level];
					permutation[level] = tmp;
				}

				is_similar_pairs_recursive(level + 1, n, zoznamy, one_pairs, other_pairs, permutation, is_similar);

				if (i!=level) {
					int tmp = permutation[i];
					permutation[i] = permutation[level];
					permutation[level] = tmp;
				}

			}

		}

	}


} 

/* is  zoznamy isomorphic to one_pairs, other_pairs*/
bool is_similar_pairs(const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs) {
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	bool is_similar = false;

	is_similar_pairs_recursive(0, n, zoznamy, one_pairs, other_pairs, permutation, is_similar);

	return is_similar;
} 

/* search for vec_zoznamy wheater it contains isomorphic (one, other) pairs to one_pairs and other_pairs */
bool is_similar_vec_pairs(const int n, const vector<Zoznamy> &vec_zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs) {
	for (vector<Zoznamy>::const_iterator it_vec = vec_zoznamy.begin(); it_vec != vec_zoznamy.end(); ++it_vec) {

		if (is_similar_pairs(n, *it_vec, one_pairs, other_pairs))
			return true;
		
	}

	return false;

} 

/* reduces summable pairs with respect to commutativity property 
 * i.e. it removes (i, j), (j, i) duplicates and replaces it with 
 * just (i, j), i<=j.
 */
void reduce_summable_pairs(const Zoznamy &zoznamy, Zoznamy &reduced) {
	for (set<pair<int, int> >::const_iterator it_set = zoznamy.one_pairs.begin(); it_set != zoznamy.one_pairs.end(); ++it_set)
		if (it_set->first <= it_set->second)
			reduced.one_pairs.insert(*it_set);	

	for (set<pair<int, int> >::const_iterator it_set = zoznamy.other_pairs.begin(); it_set != zoznamy.other_pairs.end(); ++it_set)
		if (it_set->first <= it_set->second)
			reduced.other_pairs.insert(*it_set);	

}



void gen_rel(int x, int y, const int n, int (&relation)[MAX][MAX], vector<Relation> &relations) {
	if (x == n - 1) {

		if (admits_reverse_of_relation(n, relation)) {

			if (!is_isomorphic(n, relation, relations)) {

				/* disp relation */

				for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						cout << relation[i][j];
					cout << endl;
				}
				cout << "-R-"<<endl;

				relations.push_back(Relation(n, relation));

				{

//					vector< pair< set< pair<int,int> >, set< pair<int,int> > > > HHHH;
					vector<Zoznamy> HHHH;
					vector<hea::Permutation> IIII;


					vector<hea::Permutation> vec_ortho;

					get_vector_of_orthosupplement_candidates(n, relation, vec_ortho);

					for (vector<hea::Permutation>::const_iterator it = vec_ortho.begin(); it != vec_ortho.end(); ++it) {

					set<pair<int, int> > one_pairs;

					for (int i = 0; i < n; i++) {
						one_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[i]));
						if (i != it->permutation[i])
							one_pairs.insert(pair<int, int>(it->permutation[i] + 1, 1 + i));

					}

/*
					if (one_pairs.size() == (unsigned int) n) {
*/

					set<pair<int, int> > other_pairs;
					
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						if (i != j && relation[i][j]) {

							other_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[j]));
							if (i != it->permutation[j])

								other_pairs.insert(pair<int, int>(1 + it->permutation[j], i + 1));
						}




					if (!is_similar_vec_pairs(n, HHHH, one_pairs, other_pairs)) {
					
						IIII.push_back(hea::Permutation(*it));
						HHHH.push_back(Zoznamy(one_pairs, other_pairs));

					}

				}
/*
			cout << "IIII size: " << IIII.size() << endl;
			cout << "HHHH size: " << HHHH.size() << endl;
*/
			/* display candidates */
			cout << "----------------------------" << endl;
			vector<hea::Permutation>::const_iterator it_perm = IIII.begin();
			vector<Zoznamy>::const_iterator it_zozn = HHHH.begin();
			while (it_perm != IIII.end()) {
				cout << "Candidate: ";
				for (int i = 0; i < n; i++)
					cout << 1 + it_perm->permutation[i]; 
				cout << endl;

				cout << "Candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;



				Zoznamy reduced;
				reduce_summable_pairs(*it_zozn, reduced);

				set<Candidate> candidates;

				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					candidates.insert(Candidate(*it_set, true));

				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 

					candidates.insert(Candidate(*it_set, false));

				/* insert pairs (0, x) into candidates */
				for (int i = 1; i <=n; i++)
					candidates.insert(Candidate(pair<int,int>(0, i), false));
				candidates.insert(Candidate(pair<int, int>(0, n + 1), true));


				/* display candidates */

				cout << "Candidates: ";

				for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
					
					cout << "(" << it_cand->candidate_pair.first << ", " << it_cand->candidate_pair.second << ")";
					if (it_cand->is_one)
						cout << "*";

					cout << ", ";

				}
				cout << endl;
				

				/* candidates for 0 + x */
				{
					for (int i = 0; i <= n - 1; i++) {
						cout << "0 + " ;
						hea::disp_elem(cout, n + 2, i + 1);
						cout << " is subset of {";
						set<int> set0xcand;
						set0xcand.insert(0);

						for (int j = 0; j < n; j++)
							if (relation[j][i]) {
#ifdef DEBUG
								cout << j + 1 << " <= " << i + 1 << endl;
#endif
								if (j != i) {
									/* if i + k exists then j + k must exist */
									bool is_OK = true;
									for (int k = 1; k <= n; k++)
										if (candidates.find(Candidate(i + 1, k)) != candidates.end()  && 
											candidates.find(Candidate(j + 1, k)) == candidates.end()) {
#ifdef DEBUG
											cout << "Failed: " << i + 1 << " + " << k << " exists, but " << j + 1 << " + " << k << " don't." << endl;
#endif
											is_OK = false;
											break;
										}
									if (!is_OK) continue;
								}
								set0xcand.insert(j + 1);
							}


						hea::disp_set(cout, n + 2, set0xcand);
						cout << "}";
						cout << endl;
					}

				}
				/* */
/* is_summable array */
				
				int arr_is_summable[MAX][MAX];
				{
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						arr_is_summable[i][j] = 0;

					for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) {
						arr_is_summable[it_set->first - 1][it_set->second - 1] = 1;
						arr_is_summable[it_set->second - 1][it_set->first - 1] = 1;
						
					}

					for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) {
						arr_is_summable[it_set->first - 1][it_set->second - 1] = 1;
						arr_is_summable[it_set->second - 1][it_set->first - 1] = 1;
						
					}
					cout << "Is summable array: " << endl;
					for (int i = 0; i < n; i++) {
						for (int j = 0; j < n; j++)
							cout << arr_is_summable[i][j];
						cout << endl;
					}
				}

/* negative consequences of associativity */
				map<pair<int,int>, set<int> > not_subset_of_map;
				{
					  for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) {
						if (it_set->first <= it_set->second) {
							cout << "(";
							hea::disp_elem(cout, n + 2, it_set->first);
							cout << ", ";
							hea::disp_elem(cout, n + 2, it_set->second);
							cout  << ") is not subset of: ";
#if 0
							if (it_set->first == it_set->second) {
								set<int> num_sets;
								for (int i = 0; i < n; i++)
									if (i == it_perm->permutation[it_set->first - 1] || !arr_is_summable[i][it_set->first - 1]) {
/*
										cout << "\t";
										hea::disp_elem(cout, n + 2, i + 1);
										cout << " is not summable with "; 
										hea::disp_elem(cout, n + 2, it_set->first); 
										cout << endl;
*/
										int elem_num_set = 1; /* 0 is in set */
										cout << "{0";
										for (int j = 0; j < n; j++)
											if (arr_is_summable[i][j]) {
												elem_num_set |= (1 << (j + 1));
												cout << ", ";
												hea::disp_elem(cout, n + 2, j + 1);						
											}
										num_sets.insert(elem_num_set);
										cout << "}, ";
										//cout << elem_num_set << endl;
									}
								not_subset_of_map.insert(pair<pair<int, int>, set<int> >(*it_set, num_sets));
								cout << endl;

							}
							else {
#endif
							set<int> num_sets;
							for (int i = 0; i < n; i++)
								if (i == it_perm->permutation[it_set->first - 1] || i == it_perm->permutation[it_set->second - 1] || !arr_is_summable[i][it_set->first - 1] || !arr_is_summable[i][it_set->second - 1]) {
/*
									cout << "\t";
									hea::disp_elem(cout, n + 2, i + 1);
									cout << " is not summable with "; 
									hea::disp_elem(cout, n + 2, it_set->first); 
									cout << endl;
*/
									int elem_num_set = 1; /* 0 is in set */
									cout << "{0";
									for (int j = 0; j < n; j++)
										if (arr_is_summable[i][j]) {
											elem_num_set |= (1 << (j + 1));
											cout << ", ";
											hea::disp_elem(cout, n + 2, j + 1);						
										}
									num_sets.insert(elem_num_set);
									cout << "}, ";
									//cout << elem_num_set << endl;
								}
							not_subset_of_map.insert(pair<pair<int, int>, set<int> >(*it_set, num_sets));
							cout << endl;
							set<int> reduced_num_sets;
							for (set<int>::const_iterator it_num_sets = num_sets.begin(); it_num_sets != num_sets.end(); ++it_num_sets) {
								bool insert = true;
								set<int>::const_iterator it_num_sets_pom = it_num_sets;
								for (; it_num_sets_pom != num_sets.end(); ++it_num_sets_pom) {
									if (

										(it_num_sets != it_num_sets_pom && (*it_num_sets & *it_num_sets_pom) == *it_num_sets) || *it_num_sets == 1 + (1 << it_set->first) || *it_num_sets == 1 + (1 << it_set->second)) {
										insert = false;
										break;
									}
								}
								if (insert)
									reduced_num_sets.insert(*it_num_sets);
							}
							not_subset_of_map[*it_set] = reduced_num_sets;


							/* disp reduced */
							cout << "Reduced list: ";
							for (set<int>::const_iterator it_reduced = reduced_num_sets.begin(); it_reduced != reduced_num_sets.end(); ++it_reduced) {
								cout << "{0";
								for (int j = 1; j <=n; j++)
									if (*it_reduced & (1 << j)) {
										cout << ", ";
										hea::disp_elem(cout, n + 2, j);
									}
								cout << "}, ";
								
							}	
							cout << endl;

#if 0
							}	
#endif

//							cout << endl;					
					}
				  }
				
				}
				/* display map */
				
/*
				cout << "Reduced candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Reduced candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;
*/

				cout << "----------------------------" << endl;


				++it_perm;
				++it_zozn;
			}

			}


		}

		}
	}
	else {
		for (int i = 0; i < 3; i++) {
			switch (i) {
				case 0:
					relation[x][y] = 0;
					relation[y][x] = 0;
					break;
				case 1:
					relation[x][y] = 1;
					relation[y][x] = 0;
					break;
				case 2:
					relation[x][y] = 0;
					relation[y][x] = 1;
					break;
			}

			if (y < n - 1)
				gen_rel(x, y + 1, n, relation, relations);
			else
				gen_rel(x + 1, x + 2, n, relation, relations);
		}
	}
}


int main(int argc, char** argv) {

	if (argc < 2) {
		cout << "Lists all cases to generate models of hyper effect algebras for size n,"  << endl
		    <<  "i.e. reflexive and antisymmetric relations with orthosupplement candidates.\n" << endl
		    << argv[0] << " <n>" << endl<< endl
		    << "<n> = size of the model " << MAX + 2 << " >= n > 2" << endl

		    << endl << "Interpretation of relation: " << endl
			<< "\t R(i, j) = 1, i is in relation with j" << endl
			<< "\t R(i, j) = 0, i is not in relation with j" << endl
			<< "for n-2 >= i >= 1 and n-2 >= j >=1." << endl; 

		
		return 0;
	}

	int relation[MAX][MAX];
	const int n = atoi(argv[1]) - 2;
	
	if (n <= 0) {
		cout << "<n> must be greater than 2." << endl;
		return 0;
	}
	if (n > MAX) {
		cout << "<n> must be less than or equal to " << MAX << "." << endl
		<< "If necessary, recompile sources with greater setting of MAX constant in consts.h file." << endl;
		return 0;
	}


	vector<Relation> relations;

	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		relation[i][j] = 0;

	for (int i = 0; i < n; i++)
		relation[i][i] = 1;

	gen_rel(0, 1, n, relation, relations);

/*
	map<pair<int, int>, bool> is_last_of_x;
	is_last_of_x.insert(pair<pair<int, int>, bool>(pair<int, int>(1,2), true));
*/

}
