#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <fstream>
#include <vector>

#include "hyperEA3_common.h"
#include "hea.h"
#include "hyperEA3_parse.h"
	

using namespace std;



struct Permutation {
	int n;
	int permutation[MAX];

	Permutation(const int _n, const int (&_permutation)[MAX]);
};

Permutation::Permutation(const int _n, const int (&_permutation)[MAX]) {
	n = _n;
	for (int i = 0; i < n; i++)
		permutation[i] = _permutation[i];	
}




void disp_set(const int n, const set<int> &ds) {
/*
	for (set<int>::iterator it = ds.begin(); it != ds.end(); ++it) {
		cout << *it << ", ";
	}
//	cout << endl;
*/

	set<int>::iterator it = ds.begin();
	if (it != ds.end()) {
		//cout << *it;
		hea::disp_elem(cout, n, *it);
		++it;
		do {
			if (it != ds.end()) {
				cout << ", "; 
				hea::disp_elem(cout, n, *it);
				++it;
			}
		} while (it != ds.end());
	}

}

void generate_all_order_preserving_permutations(int level, const hea::Hea &hea, int (&permutation)[MAX], vector<Permutation> &vec_order_preserve) {

	if (level == hea.n - 2) {
		
		for (int i = 1; i < hea.n - 1; i++)
		for (int j = 1; j < hea.n - 1; j++) {
			if (hea::is_summable(hea.op, i, hea.complement[j]) && !hea::is_summable(hea.op, permutation[i], hea.complement[permutation[j]]))
				return ;
			if (hea::is_summable(hea.op, permutation[i], hea.complement[permutation[j]]) && !hea::is_summable(hea.op, i, hea.complement[j]))
				return ;
		}

		vec_order_preserve.push_back(Permutation(hea.n, permutation));


	}
	else {
		for (int i = level; i < hea.n - 1; i++) {
			//swap level i
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
			generate_all_order_preserving_permutations( level + 1, hea, permutation, vec_order_preserve);
			//swap level i
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
		}
		
	}

}

int main(int argc, char **argv) {

	if (argc < 2) {
		cout << "Usage: " << argv[0] << " filename [filename2 ...]" << endl;
		return 0;
	
	}

	vector<hea::Hea> non_iso_list;
//	bool is_first = true;

	const int n = 6;

	set<int> setCandidates[1 << MAX];
	const set<int>* op[MAX][MAX];

	// init
	for (int i = 0; i < 1 << n; i++) {
		setCandidates[i] = set<int>();
		hea::init_set_i(setCandidates[i], i);
	}
	
	int complement[MAX];
	int idx = 1;
	while (idx < argc) {
		ifstream ifs;

		ifs.open(argv[idx], ifstream::in);
		
		if (ifs.good()) {
			cerr << "Loading file " << argv[idx] << "." << endl;

	/*
			// skip header and copy header
			
			int to_skip = 2;
			char buffer[1024];
			while (to_skip > 0) {
				ifs.getline(buffer, 1024);
				if (!strncmp(buffer, "-----", 5)) {
					//cout << "Skipped..." <<endl;
					to_skip--;
				}
				cout << buffer<<endl;
			}

	*/

			while(hea::parse_model(ifs, n, op, complement, setCandidates)) {

				hea::Hea original_hea(n, op, complement);
				
				// modify op - remove all zeros from sets, except 0 + 0
				//for (int i = 0 ; i < n; i++)
				// modify op - remove all zeros from sets, except 0 + x
				for (int i = 1 ; i < n; i++)
				for (int j = i ; j < n; j++) {
					if (i != 0 || j != 0) {

						int idx = op[i][j] - &setCandidates[0];
//						cout << "i: " << i << ", j: " << j << ", idx: " << idx << endl;
						
						if (idx == 0)
							continue;

						idx = (idx / 2) * 2;

//						cout << "i: " << i << ", j: " << j << ", idxM: " << idx << endl;

						op[i][j] = &setCandidates[idx];
						if (i != j)
							op[j][i] = &setCandidates[idx];
					}

				}
					
				if (!hea::is_associative(n, op)) {
					cout << "Original: " << endl;
					hea::disp_hea(cout, original_hea);
				

					cout << "Modified non associative: " << endl;
					hea::Hea tested_hea(n, op, complement);
					hea::disp_hea(cout, tested_hea);
					return 0;
				}




				// skip one line containing -M-
				char buffer[1024];
				ifs.getline(buffer, 1024);
			}
				
		}
		else {
			cerr << "Cannot open or read from file '" << argv[idx] << "'." << endl;
		}
		ifs.close();

		idx++;
	}

	return 0;
}
